from .Photo import Photo
from math import sqrt
from itertools import combinations
from statistics import mean, stdev


class Photoset:
    """Class to perform likeness calculations on multiple Photo objects."""

    def __init__(self, photolist: []):
        """Constructor

        Parameters:
        photolist ([]): list containing 2 or more Photo objects

        """

        self.__saturation_likeness = None
        self.__overall_likeness = None
        self.__lightness_likeness = None
        self.__color_likeness = None
        self.__fraction_likeness = None
        self.__avg_lightness = None
        self.__avg_saturation = None

        if len(photolist) < 2:
            raise ValueError("Length of input photo list must be at least 2.")
        for item in photolist:
            if type(item) != Photo:
                raise TypeError("Non-Photo object found in input list")
        self.photolist = photolist
        self.filenames_index = {}
        for photo in photolist:
            self.filenames_index[photo.filename] = photolist.index(photo)
        self.length = len(photolist)

    @staticmethod
    def __color_difference(col1, col2):
        sum_of_sq = 0
        for i in range(3):
            sum_of_sq += (col1[i] - col2[i]) ** 2
        return sqrt(sum_of_sq) / (sqrt(3 * (255 ** 2)))

    def __get_avg_rgb(self):
        averages = {}
        colorlist = ['black', 'blue', 'cyan', 'dark grey', 'green', 'grey', 'light grey', 'magenta', 'orange',
                     'red', 'turquois', 'violet', 'white', 'yellow']
        for color in colorlist:
            averages[color] = []
        for photo in self.photolist:
            for color in photo.get_color_data():
                averages[color].append(photo.get_color_data()[color]['RGB'])
        for color in averages:
            while len(averages[color]) < self.length:
                averages[color].append((0, 0, 0))  # accounts for photos lacking a specific color
        return averages

    def __get_fraction_rgb(self):
        fractions = {}
        colorlist = ['black', 'blue', 'cyan', 'dark grey', 'green', 'grey', 'light grey', 'magenta', 'orange',
                     'red', 'turquois', 'violet', 'white', 'yellow']
        for color in colorlist:
            fractions[color] = []
        for photo in self.photolist:
            for color in photo.get_color_data():
                fractions[color].append(round(photo.get_color_data()[color]['%'], 2))
        for color in fractions:
            while len(fractions[color]) < self.length:
                fractions[color].append(0)
        return fractions

    def get_color_likeness(self):
        """Get the likeness of Photo objects in terms of color shades."""

        if self.__color_likeness is not None:
            return self.__color_likeness

        data = self.__get_avg_rgb()
        running_sum = 0
        for color in data:
            differences = [self.__color_difference(c1, c2) for c1, c2 in combinations(data[color], 2)]
            running_sum += sum(differences) / len(differences)
        x = running_sum / len(data)
        self.__color_likeness = round(1 - abs((1 / (1 + 2.7 ** (-4 * x)) - .5)) * 2, 4)
        return self.__color_likeness

    def get_fraction_likeness(self):
        """Get the likeness of Photo objects in terms of fractional makeup of colors."""

        if self.__fraction_likeness is not None:
            return self.__fraction_likeness

        data = self.__get_fraction_rgb()
        running_sum = 0
        for color in data:
            differences = [abs(p1 - p2) for p1, p2 in combinations(data[color], 2)]
            running_sum += sum(differences) / len(differences)
        x = running_sum / len(data)
        self.__fraction_likeness = round(1 - abs((1 / (1 + 2.7 ** (-20 * x)) - .5)) * 2, 4)
        return self.__fraction_likeness

    def get_lightness_likeness(self):
        """Get the likeness of Photo objects in terms of lightness."""

        if self.__lightness_likeness is not None:
            return self.__lightness_likeness

        lightness_list = []
        for photo in self.photolist:
            lightness_list.append(photo.get_average_lightness())
        differences = [abs(x - y) for x, y in combinations(lightness_list, 2)]
        self.__lightness_likeness = round(1 - sum(differences) / len(differences), 4)
        return self.__lightness_likeness

    def get_saturation_likeness(self):
        """Get the likeness of Photo objects in terms of saturation."""

        if self.__saturation_likeness is not None:
            return self.__saturation_likeness

        saturation_list = []
        for photo in self.photolist:
            saturation_list.append(photo.get_average_saturation())
        differences = [abs(x - y) for x, y in combinations(saturation_list, 2)]
        self.__saturation_likeness = round(1 - sum(differences) / len(differences), 4)
        return self.__saturation_likeness

    def get_overall_likeness(self, color_weight=.25, fraction_weight=.25, lightness_wieght=.25,
                             saturation_weight=.25):
        """Get the likeness of Photo objects overall.

        Calculates overall likeness using a weighted sum of color, fraction, lightness, and saturation likeness values.

        Parameters:
        color_weight (float): weight attributed to color
        fraction_weight (float): weight attributed to fractional color makeup
        lightness_weight (float): weight attributed to lightness
        saturation_weight (float): weight attributed to saturation

        """

        if self.__overall_likeness is not None:
            return self.__overall_likeness

        if color_weight + fraction_weight + saturation_weight + lightness_wieght != 1.0:
            raise ValueError("Wieghts must add up to 1.")

        self.__overall_likeness = round(
            color_weight * self.get_color_likeness() + fraction_weight * self.get_fraction_likeness() +
            lightness_wieght * self.get_lightness_likeness() +
            saturation_weight * self.get_saturation_likeness(), 4)
        return self.__overall_likeness

    def get_avg_lightness(self):
        """Get the average lightness of photos in a Photoset."""

        if self.__avg_lightness is not None:
            return self.__avg_lightness

        lightnesses = []
        for p in self.photolist:
            lightnesses.append(p.get_average_lightness())
        self.__avg_lightness = mean(lightnesses)
        return self.__avg_lightness

    def get_std_lightness(self):
        """Get the standard deviation of lightness of photos in a Photoset."""

        lightnesses = []
        for p in self.photolist:
            lightnesses.append(p.get_average_lightness())
        return stdev(lightnesses)

    def get_avg_saturation(self):
        """Get the average saturation of photos in a Photoset."""

        if self.__avg_saturation is not None:
            return self.__avg_saturation

        saturations = []
        for p in self.photolist:
            saturations.append(p.get_average_saturation())
        self.__avg_saturation = mean(saturations)
        return self.__avg_saturation

    def get_std_saturation(self):
        """Get the standard deviation of saturation of photos in a Photoset."""

        saturations = []
        for p in self.photolist:
            saturations.append(p.get_average_saturation())
        return stdev(saturations)



from colorsys import rgb_to_hls
from PIL import Image
from math import sqrt
from collections import Counter
from statistics import stdev, median, StatisticsError


class Photo:
    """Class that stores color data for a specific photo."""

    def __init__(self, filename, pixel_count=30000):
        """Constructor.

        Parameters:
        filename (String): path to photo file
        pixel_count (int): number of pixels that should be loaded to be used by class

        """
        self.filename = filename
        self.im = Image.open(filename)
        self.im = self.im.resize([int(self.__resize_scale(self.im.size, float(pixel_count)) * s) for s in self.im.size],
                                 Image.ANTIALIAS)
        self.pixel_data = list(self.im.getdata())
        self.num_pixels = len(self.pixel_data)
        self.__color_stats = {}
        self.__color_data = self.__get_average_shades(self.__sort_colors(self.pixel_data))

    @staticmethod
    def __average_shade(pixel_list):
        if len(pixel_list) == 0:
            return None
        r, g, b = 0, 0, 0
        for pixel in pixel_list:
            r += pixel[0]
            g += pixel[1]
            b += pixel[2]
        return round(r / len(pixel_list)), round(g / len(pixel_list)), round(b / len(pixel_list))

    @staticmethod
    def __resize_scale(size, pixels):
        x, y = size
        return sqrt(pixels / (x * y))

    @staticmethod
    def __rgb_to_hsl(r, g, b):
        r, g, b = r / 255.0, g / 255.0, b / 255.0  # normalize rgb values
        h, l, s = rgb_to_hls(r, g, b)
        return h * 360.0, s, l

    def __color_classifier(self, tup):
        r, g, b = tup
        h, s, l = self.__rgb_to_hsl(r, g, b)
        if s >= .15:
            if 0 <= h < 40:
                return 'orange'
            elif 40 <= h < 80:
                return 'yellow'
            elif 80 <= h < 120:
                return 'green'
            elif 120 <= h < 160:
                return 'turquois'
            elif 160 <= h < 200:
                return 'cyan'
            elif 200 <= h < 240:
                return 'blue'
            elif 240 <= h < 280:
                return 'violet'
            elif 280 <= h < 320:
                return 'magenta'
            elif 320 <= h <= 360:
                return 'red'
        else:
            grey_cutoffs = [0.0, .25, .50, .75, 1.0]
            distances = []
            for val in grey_cutoffs:
                distances.append(abs(l - val))
            closest = grey_cutoffs[distances.index(min(distances))]
            if closest == 0.0:
                return "black"
            elif closest == .25:
                return "dark grey"
            elif closest == .50:
                return "grey"
            elif closest == .75:
                return "light grey"
            elif closest == 1.0:
                return "white"

    def __sort_colors(self, pixel_list):
        color_dict = {'red': [], 'orange': [], 'yellow': [],
                      'green': [], 'turquois': [], 'cyan': [],
                      'magenta': [], 'blue': [], 'violet': [],
                      'black': [], 'dark grey': [], 'grey': [],
                      'light grey': [], 'white': []}
        for pixel in pixel_list:
            color = self.__color_classifier(pixel)
            color_dict[color].append(pixel)
        for color in color_dict:
            avg = self.__average_shade(color_dict[color])
            self.__color_stats[color] = {}
            self.__color_stats[color]["mean"] = avg
            r = []
            g = []
            b = []
            for pixel in color_dict[color]:
                for i in range(3):
                    if i == 0:
                        r.append(pixel[i])
                    elif i == 1:
                        g.append(pixel[i])
                    else:
                        b.append(pixel[i])
            try:
                self.__color_stats[color]["median"] = (median(r), median(g), median(b))
                self.__color_stats[color]["stdev"] = (round(stdev(r), 2), round(stdev(g), 2), round(stdev(b), 2))
            except StatisticsError:
                self.__color_stats[color]["stdev"] = None
                self.__color_stats[color]["median"] = None
        return color_dict

    def __get_average_shades(self, color_dict):
        ret_dict = {}
        for color in color_dict:
            avg_shade = self.__average_shade(color_dict[color])
            if avg_shade:  # some colors are not present in an image so the avg_shade is None
                ret_dict[color] = {}
                ret_dict[color] = {'RGB': avg_shade, 'HEX': '#%02x%02x%02x' % avg_shade,
                                   '%': round(len(color_dict[color]) / self.num_pixels, 4)}
        return ret_dict

    def get_n_most_common_colors(self, n):
        """Get the 'n' most common colors in the Photo.

        Returns:
        dictionary of most common colors and their respective RGB, HEX, and percentage makeup values.
        """
        l = {}
        for color in self.__color_data:
            l[color] = self.__color_data[color]['%']
        count_dict = Counter(l)
        l = {}
        for color in count_dict.most_common(n):
            l[color[0]] = self.__color_data[color[0]]
        return l

    def get_average_lightness(self):
        """Get the average lightness of Photo.

        Returns:
        int: average of lightness of pixels using l value in HSL.
        """
        running_sum = 0
        for pixel in self.pixel_data:
            hsl = self.__rgb_to_hsl(pixel[0], pixel[1], pixel[2])
            running_sum += hsl[2]
        return running_sum / self.num_pixels

    def get_average_saturation(self):
        """Get the average saturation of Photo.

        Returns:
        int: average of saturation of pixels using s value in HSL.
        """
        running_sum = 0
        for pixel in self.pixel_data:
            hsl = self.__rgb_to_hsl(pixel[0], pixel[1], pixel[2])
            running_sum += hsl[1]
        return running_sum / self.num_pixels

    def get_color_data(self, colors: [str] = None):
        """Get color data of a Photo.

        Parameters:
        colors [str]: list of colors to be included in the returned data
        """

        if colors is None or colors == []:
            return self.__color_data
        return_data = {}
        for color in colors:
            try:
                return_data[color] = self.__color_data[color]
            except KeyError:
                pass
        return return_data

    def get_color_stats(self, colors: [str] = None):
        """Get color stats of a Photo.

        Parameters:
        colors [str]: list of colors to be included in the returned data

        Returns:
        return_data {}: contains mean shade of color, standard deviation and median of rgb components
        """

        if colors is None or colors == []:
            return self.__color_stats
        return_data = {}
        for color in colors:
            try:
                return_data[color] = self.__color_stats[color]
            except KeyError:
                pass
        return return_data



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="colorstats",
    version="1.1.2",
    author="Nicholas Ramkissoon",
    description="Package for processing color data from images",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nramkissoon/ColorStats",
    packages=setuptools.find_packages(),
    python_requires='>=3',
    classifiers=[
        'Intended Audience :: Developers',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)